import warnings
import numpy as np
from dataclasses import dataclass, field
from enum import Enum

_WIDTH = 75

class StopCrit(Enum):
    NOT_OPTIMIZED = -1
    INTERRUPTED = 0
    CONVERGED = 1

    @property
    def has_converged(self):
        return (self.value > 0)

@dataclass
class TypeValidator:
    def validate_type(self):
        for f_name, f_def in self.__dataclass_fields__.items():
            f_type = type(getattr(self, f_name))
            if not (f_type == list or f_type == np.ndarray):
                if not issubclass(f_type,f_def.type):
                    raise ValueError(f"Expected {f_def.type} for {f_name}")

@dataclass
class ProjgradParams(TypeValidator):
    """
    Parameters for the method projgrad in srenet.projgrad.
    """
    maxit       : int       = 10_000       # Maximum number of iterations
    tol         : float     = 1e-15        # Tolerance required on the gap
    verbosity   : bool      = True         # Toogle verbosity of the method
    showevery   : int       = 10           # Show iter info every `showevery`
    screen      : bool      = False        # Toogle screening tests
    relax       : bool      = False        # Toogle relaxing tests
    coherence   : bool      = False        # Compute coherence every iterations
    Sinit       : list[int] = field(default_factory=list) # Init. knowledge of S
    def validate(self):
        self.validate_type()

@dataclass
class ProjgradMonitor():
    """
    Monitoring values of the method projgrad in srenet.projgrad.
    """
    obj         : list[float] = field(default_factory=list) # Objective trace
    gap         : list[float] = field(default_factory=list) # Gap trace
    flops       : list[int]   = field(default_factory=list) # FLOPs trace
    stopcrit    : StopCrit    = StopCrit.NOT_OPTIMIZED      # Stopping criterion
    solvetime   : float       = -1                          # Solvetime
    sc          : set[int]    = field(default_factory=set)  # Screened coefs
    rc          : set[int]    = field(default_factory=set)  # Relaxed coefs
    sc_size     : list[int]   = field(default_factory=list) # Nb screened coefs over iters.
    rc_size     : list[int]   = field(default_factory=list) # Nb relaxed coefs over iters.
    coherence   : list[float] = field(default_factory=list) # Coherence trace

def validate_inputs(x0, A, y, l1, l2):
    if not issubclass(type(x0),np.ndarray):
        raise ValueError("Expected np.ndarray for x0")
    if not issubclass(type(A),np.ndarray):
        raise ValueError("Expected np.ndarray for A")
    if not issubclass(type(y),np.ndarray):
        raise ValueError("Expected np.ndarray for y")
    if y.shape[0] != A.shape[0]:
        raise ValueError("Dimension missmatch between A and y")
    if x0.shape[0] != A.shape[1]:
        raise ValueError("Dimension missmatch between A and x0")
    if not issubclass(type(l1),float):
        raise ValueError("Expected float for l1")
    if not issubclass(type(l2),float):
        raise ValueError("Expected float for l2")
    if l1 <= 0:
        raise ValueError("Expected a positive value for l1")
    if l2 <= 0:
        raise ValueError("Expected a positive value for l2")
    if not np.allclose(np.linalg.norm(y,2),1):
        raise ValueError("Expected y to be normalized")
    for i in range(A.shape[1]):
        if not np.allclose(np.linalg.norm(A[:,i],2),1):
            raise ValueError("Expected columns of A to be normalized")
    
def print_header(A, l1, l2, params):
    print('\n'+'*'*_WIDTH)
    print(("Projected-Gradient for the Elastic-Net problem\n").center(_WIDTH))
    print("Parameters")
    print("  dim    : {:d}x{:d}".format(A.shape[0],A.shape[1]))
    print("  l1     : {:.2e}".format(l1))
    print("  l2     : {:.2e}".format(l2))
    print("  maxit  : {:d}".format(params.maxit))
    print("  tol    : {:.2e}".format(params.maxit))
    print("  screen : {}".format(params.screen))
    print("  relax  : {}".format(params.relax))
    
    print()

    HEADER = "{}|{}{}{}{}{}".format(
        'Iter'.ljust(7),
        'Obj'.rjust(10),
        'Gap'.rjust(10),
        'Flops'.rjust(10),
        'Screen'.rjust(7),
        'Relax'.rjust(7),
    )

    print(HEADER)
    print("-"*len(HEADER))
    
def print_trace(it, monitor):
    TRACE = "{}|{}{}{}{}{}".format(
        '{:d}'.format(it).ljust(7),
        '{:.2e}'.format(monitor.obj[it]).rjust(10),
        '{:.2e}'.format(monitor.gap[it]).rjust(10),
        '{:.2e}'.format(monitor.flops[it]).rjust(10),
        '{:d}'.format(monitor.sc_size[it]).rjust(7),
        '{:d}'.format(monitor.rc_size[it]).rjust(7),
    )
    print(TRACE)

def print_footer(monitor):
    print()
    if not monitor.stopcrit.has_converged:
        warnings.warn("Algorithm did not converge")
    print("Stop crit  : {}".format(monitor.stopcrit.name))
    print("Obj value  : {:.5e}".format(monitor.obj[-1]))
    print("Final gap  : {:.5e}".format(monitor.gap[-1]))
    print("Flops      : {:.5e}".format(monitor.flops[-1]))
    print("Solve time : {:.2f} seconds".format(monitor.solvetime))
    print()
    print('*'*_WIDTH + '\n')